from pydantic import Field
from rath.links.file import FileExtraction
from rath.links.dictinglink import DictingLink
from rath.links.auth import AuthTokenLink


def init_services(service_builder_registry):
    from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
    from rath.links.split import SplitLink
    from rath.contrib.fakts.links.graphql_ws import FaktsGraphQLWSLink
    from rath.contrib.herre.links.auth import HerreAuthLink
    from fakts import Fakts
    from herre import Herre
    from arkitekt_next.service_registry import Params
    from arkitekt_next.model import Requirement

    from mikro_next.mikro_next import MikroNext
    from mikro_next.rath import MikroNextLinkComposition, MikroNextRath
    from rath.links.split import SplitLink
    from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
    from rath.contrib.fakts.links.graphql_ws import FaktsGraphQLWSLink
    from rath.contrib.herre.links.auth import HerreAuthLink
    from mikro_next.contrib.fakts.datalayer import FaktsDataLayer
    from mikro_next.links.upload import UploadLink
    from mikro_next.datalayer import DataLayer
    from graphql import OperationType
    from herre import Herre
    from fakts import Fakts

    from arkitekt_next.model import Manifest

    try:
        from rekuest_next.links.context import ContextLink
        from rath.links.compose import TypedComposedLink

        class ArkitektMikroNextLinkComposition(TypedComposedLink):
            fileextraction: FileExtraction = Field(default_factory=FileExtraction)
            """ A link that extracts files from the request and follows the graphql multipart request spec"""
            dicting: DictingLink = Field(default_factory=DictingLink)
            """ A link that converts basemodels to dicts"""
            upload: UploadLink
            """ A link that uploads supported data types like numpy arrays and parquet files to the datalayer"""
            auth: AuthTokenLink
            """ A link that adds the auth token to the request"""
            """ A link that splits the request into a http and a websocket request"""
            assignation: ContextLink = Field(default_factory=ContextLink)
            split: SplitLink

    except ImportError:
        ArkitektMikroNextLinkComposition = MikroNextLinkComposition

    class ArkitektMikroNextRath(MikroNextRath):
        link: ArkitektMikroNextLinkComposition

    class ArkitektNextMikroNext(MikroNext):
        rath: ArkitektMikroNextRath
        datalayer: DataLayer

    def builder_mikro(fakts: Fakts, herre: Herre, params: Params, manifest: Manifest):
        datalayer = FaktsDataLayer(fakts_group="datalayer", fakts=fakts)

        return ArkitektNextMikroNext(
            rath=ArkitektMikroNextRath(
                link=ArkitektMikroNextLinkComposition(
                    auth=HerreAuthLink(herre=herre),
                    upload=UploadLink(
                        datalayer=datalayer,
                    ),
                    split=SplitLink(
                        left=FaktsAIOHttpLink(
                            fakts_group="mikro", fakts=fakts, endpoint_url="FAKE_URL"
                        ),
                        right=FaktsGraphQLWSLink(
                            fakts_group="mikro", fakts=fakts, ws_endpoint_url="FAKE_URL"
                        ),
                        split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                    ),
                )
            ),
            datalayer=datalayer,
        )

    def fake_builder(fakts, herre, params, manifest):
        return FaktsDataLayer(fakts_group="datalayer", fakts=fakts)

    service_builder_registry.register(
        "mikro",
        builder_mikro,
        Requirement(
            service="live.arkitekt.mikro",
            description="An instance of ArkitektNext Mikro to make requests to the user's data",
            optional=True,
        ),
    )
    service_builder_registry.register(
        "datalayer",
        fake_builder,
        Requirement(
            service="live.arkitekt.s3",
            description="An instance of ArkitektNext Datalayer to make requests to the user's data",
            optional=True,
        ),
    )



import asyncio

from mikro_next.scalars import ArrayLike, ParquetLike, FileLike
from rath.links.parsing import ParsingLink
from rath.operation import Operation, opify
from mikro_next.io.types import Uploader
from typing import Any
from mikro_next.io.upload import aupload_bigfile, aupload_xarray, aupload_parquet
from pydantic import Field
from concurrent.futures import ThreadPoolExecutor
import uuid
from functools import partial
from mikro_next.datalayer import DataLayer


async def apply_recursive(func, obj, typeguard):
    """
    Recursively applies an asynchronous function to elements in a nested structure.

    Args:
        func (callable): The asynchronous function to apply.
        obj (any): The nested structure (dict, list, tuple, etc.) to process.
        typeguard (type): The type of elements to apply the function to.

    Returns:
        any: The nested structure with the function applied to elements of the specified type.
    """
    if isinstance(
        obj, dict
    ):  # If obj is a dictionary, recursively apply to each key-value pair
        return {k: await apply_recursive(func, v, typeguard) for k, v in obj.items()}
    elif isinstance(obj, list):  # If obj is a list, recursively apply to each element
        return await asyncio.gather(
            *[apply_recursive(func, elem, typeguard) for elem in obj]
        )
    elif isinstance(
        obj, tuple
    ):  # If obj is a tuple, recursively apply to each element and convert back to tuple
        return tuple(
            await asyncio.gather(
                *[apply_recursive(func, elem, typeguard) for elem in obj]
            )
        )
    elif isinstance(obj, typeguard):  # If obj matches the typeguard, apply the function
        return await func(obj)
    else:  # If obj is not a dict, list, tuple, or matching the typeguard, return it as is
        return obj


async def afake_upload(xarray: ArrayLike, *args, **kwargs) -> str:
    return str(uuid.uuid4())


class UploadLink(ParsingLink):
    """Data Layer Upload Link

    This link is used to upload  supported types to a DataLayer.
    It parses queries, mutatoin and subscription arguments and
    uploads the items to the DataLayer, and substitures the
    DataFrame with the S3 path.

    Args:
        ParsingLink (_type_): _description_


    """

    parquet_uploader: Uploader = aupload_parquet
    xarray_uploader: Uploader = aupload_xarray
    bigfile_uploader: Uploader = aupload_bigfile
    datalayer: DataLayer

    executor: ThreadPoolExecutor = Field(
        default_factory=lambda: ThreadPoolExecutor(max_workers=4), exclude=True
    )
    _executor_session: Any = None

    async def __aenter__(self):
        self._executor_session = self.executor.__enter__()

    async def aget_image_credentials(self, key, datalayer) -> Any:
        from mikro_next.api.schema import RequestUploadMutation

        operation = opify(
            RequestUploadMutation.Meta.document,
            variables={"key": key, "datalayer": datalayer},
        )

        async for result in self.next.aexecute(operation):
            return RequestUploadMutation(**result.data).request_upload

    async def aget_table_credentials(self, key, datalayer) -> Any:
        from mikro_next.api.schema import RequestTableUploadMutation

        operation = opify(
            RequestTableUploadMutation.Meta.document,
            variables={"key": key, "datalayer": datalayer},
        )

        async for result in self.next.aexecute(operation):
            return RequestTableUploadMutation(**result.data).request_table_upload

    async def aget_bigfile_credentials(self, key, datalayer) -> Any:
        from mikro_next.api.schema import RequestFileUploadMutation

        operation = opify(
            RequestFileUploadMutation.Meta.document,
            variables={"key": key, "datalayer": datalayer},
        )

        async for result in self.next.aexecute(operation):
            return RequestFileUploadMutation(**result.data).request_file_upload

    async def aupload_parquet(
        self, datalayer: "DataLayer", parquet_input: ParquetLike
    ) -> str:
        assert datalayer is not None, "Datalayer must be set"
        endpoint_url = await datalayer.get_endpoint_url()

        credentials = await self.aget_table_credentials(parquet_input.key, endpoint_url)
        return await self.parquet_uploader(
            parquet_input,
            credentials,
            datalayer,
            self._executor_session,
        )

    async def aupload_xarray(self, datalayer: "DataLayer", xarray: ArrayLike) -> str:
        assert datalayer is not None, "Datalayer must be set"
        endpoint_url = await datalayer.get_endpoint_url()

        credentials = await self.aget_image_credentials(xarray.key, endpoint_url)
        return await self.xarray_uploader(
            xarray,
            credentials,
            datalayer,
            self._executor_session,
        )

    async def aupload_bigfile(self, datalayer: "DataLayer", file: FileLike) -> str:
        assert datalayer is not None, "Datalayer must be set"
        endpoint_url = await datalayer.get_endpoint_url()

        credentials = await self.aget_bigfile_credentials(file.key, endpoint_url)
        return await self.bigfile_uploader(
            file,
            credentials,
            datalayer,
            self._executor_session,
        )

    async def aparse(self, operation: Operation) -> Operation:
        """Parse the operation (Async)

        Extracts the DataFrame from the operation and uploads it to the DataLayer.

        Args:
            operation (Operation): The operation to parse

        Returns:
            Operation: _description_
        """

        datalayer = operation.context.kwargs.get("datalayer", self.datalayer)

        operation.variables = await apply_recursive(
            partial(self.aupload_xarray, datalayer),
            operation.variables,
            ArrayLike,
        )
        operation.variables = await apply_recursive(
            partial(self.aupload_parquet, datalayer), operation.variables, ParquetLike
        )
        operation.variables = await apply_recursive(
            partial(self.aupload_bigfile, datalayer), operation.variables, FileLike
        )

        return operation

    async def adisconnect(self):
        self.executor.__exit__(None, None, None)



import rich_click as click
import subprocess
import os


directory = os.getcwd()

file_path = os.path.dirname(__file__)


v1_file = """
projects:
  mikro:
    schema: http://localhost:8080/graphql
    documents: graphql/mikro/*/**.graphql
    extensions:
      turms:
        out_dir: mikro/api
        freeze:
          enabled: true
        stylers:
          - type: turms.stylers.default.DefaultStyler
          - type: turms.stylers.appender.AppenderStyler
            append_fragment: "Fragment"
        plugins:
          - type: turms.plugins.enums.EnumsPlugin
          - type: turms.plugins.inputs.InputsPlugin
          - type: turms.plugins.fragments.FragmentsPlugin
          - type: turms.plugins.operations.OperationsPlugin
          - type: turms.plugins.funcs.FuncsPlugin
            global_kwargs:
              - type: mikro_nextrath.MikroRath
                key: rath
                description: "The mikro rath client"
            definitions:
              - type: subscription
                is_async: True
                use: mikro_nextfuncs.asubscribe
              - type: query
                is_async: True
                use: mikro_nextfuncs.aexecute
              - type: mutation
                is_async: True
                use: mikro_nextfuncs.aexecute
              - type: subscription
                use: mikro_nextfuncs.subscribe
              - type: query
                use: mikro_nextfuncs.execute
              - type: mutation
                use: mikro_nextfuncs.execute
        processors:
          - type: turms.processors.black.BlackProcessor
        scalar_definitions:
          XArrayInput: mikro_nextscalars.XArrayInput
          File: mikro_nextscalars.File
          ImageFile: mikro_nextscalars.File
          Upload: mikro_nextscalars.Upload
          ModelData: mikro_nextscalars.ModelData
          ModelFile: mikro_nextscalars.ModelFile
          ParquetInput: mikro_nextscalars.ParquetInput
          Store: mikro_nextscalars.Store
          Parquet: mikro_nextscalars.Parquet
          ID: rath.scalars.ID
          MetricValue: mikro_nextscalars.MetricValue
          FeatureValue: mikro_nextscalars.FeatureValue
        additional_bases:
          Representation:
            - mikro_nexttraits.Representation
          Table:
            - mikro_nexttraits.Table
          Omero:
            - mikro_nexttraits.Omero
          Objective:
            - mikro_nexttraits.Objective
          Position:
            - mikro_nexttraits.Position
          Stage:
            - mikro_nexttraits.Stage
          ROI:
            - mikro_nexttraits.ROI
          InputVector:
            - mikro_nexttraits.Vectorizable
"""


@click.group()
def cli():
    pass


@cli.command()
def version():
    """Shows the current version of mikro"""


@cli.command()
def generate():
    """Generates the mikro api"""
    with open(f"{file_path}/../graphql.config.yaml", "w") as f:
        f.write(v1_file)
    subprocess.run(
        ["turms", "generate", "--config", f"{file_path}/../graphql.config.yaml"]
    )
    os.remove(f"{file_path}/../graphql.config.yaml")


if __name__ == "__main__":
    cli()



from rath.links.auth import AuthTokenLink
from rath.links.compose import TypedComposedLink
from rath.links.dictinglink import DictingLink
from rath.links.file import FileExtraction
from rath.links.split import SplitLink
from rekuest.contrib.rath.set_assignation_link import SetAssignationLink
from pydantic import Field
from mikro_next.links.upload import UploadLink


class ArkitektMikroNextLinkComposition(TypedComposedLink):
    fileextraction: FileExtraction = Field(default_factory=FileExtraction)
    dicting: DictingLink = Field(default_factory=DictingLink)
    assignation_link: SetAssignationLink = Field(default_factory=SetAssignationLink)
    upload: UploadLink = Field(default_factory=UploadLink)
    auth: AuthTokenLink
    split: SplitLink



"""
This modules provides the datalayer. 

Example:

    A simple datalayer that connects to an s3 instance via access_key and secret_key.
    You can define all of the logic within the context manager

    ```python
    from mikro_next imports Datalayer

    dl = Datalayer(access_key="XXXX", secret_key="XXXX", endpoint_url="s3.amazonaws.com")

    with dl:
        print(df.fs.ls())

    ```

    Async Usage:

    ```python
    from mikro_nextdatalayer import Datalayer

    dl = Datalayer(access_key="XXXX", secret_key="XXXX", endpoint_url="s3.amazonaws.com")

    async with dl:
        print(df.fs.ls())

    ```

 
"""

import contextvars

from koil.composition import KoiledModel


current_next_datalayer = contextvars.ContextVar("current_next_datalayer", default=None)


class DataLayer(KoiledModel):
    """Implements a S3 DataLayer

    This will be used to upload and download files from S3.

    Make sure to set the access_key and secret_key and enter the context
    manager to connect to S3 (if authentication is required for the S3 instance
    and to ensure that the context is exited when the context manager is exited
    (for future cleanup purposes on other datalayers).

    Attributes:
        fs (s3fs.S3FileSystem): The filesystem object

    """

    endpoint_url: str = ""

    async def get_endpoint_url(self):
        return self.endpoint_url

    async def __aenter__(self):
        current_next_datalayer.set(self)
        return self

    async def __aexit__(self, *args, **kwargs):
        current_next_datalayer.set(None)
        return self



from .api.schema import (
    PartialAffineTransformationViewInput,
    PartialChannelViewInput,
    PartialTimepointViewInput,
    PartialOpticsViewInput,
    PartialLabelViewInput,
    ChannelFragment,
    FluorophoreFragment,
    ObjectiveFragment,
    InstrumentFragment,
    PartialRGBViewInput,
    CameraFragment,
    EraFragment,
    AntibodyFragment,
)
from .scalars import ThreeDVector, Milliseconds, FourByFourMatrix
from rath.scalars import ID
from typing import List, Optional
import numpy as np


def accesor_from_slices(slices: List[str] = None):
    """Creates an accesor from a list of slices

    Args:
        slices (List[str]): The slices

    Returns:
        str: The accesor
    """
    return {}


def view_from_pixel_size_and_position(
    scale: ThreeDVector,
    position: ThreeDVector,
    stage: ID = None,
    **kwargs,
) -> PartialAffineTransformationViewInput:
    """Creates a view from the pixel size and position

    Args:
        x_scale (float): The x scale
        y_scale (float): The y scale
        z_scale (float): The z scale
        x_position (float): The x position
        y_position (float): The y position
        z_position (float): The z position

    Returns:
        View: The view
    """
    x_scale, y_scale, z_scale = scale
    x_position, y_position, z_position = position

    affine = np.array(
        [
            [x_scale, 0, 0, x_position],
            [0, y_scale, 0, y_position],
            [0, 0, z_scale, z_position],
            [0, 0, 0, 1],
        ]
    )

    FourByFourMatrix.validate(affine)

    return PartialAffineTransformationViewInput(
        affineMatrix=affine, stage=stage, **kwargs
    )


def view_from_channel(
    channel: ChannelFragment, c: int = None, **kwargs
) -> PartialChannelViewInput:
    """Creates a view from the channel name

    Args:
        channel_name (str): The channel name

    Returns:
        View: The view
    """
    if c is not None:
        return PartialChannelViewInput(channel=channel, cMax=c, cMin=c, **kwargs)

    return PartialChannelViewInput(channel=channel, **kwargs)


def opticsview_from_path(
    objective: ObjectiveFragment = None,
    camera: CameraFragment = None,
    instrument: InstrumentFragment = None,
    **kwargs,
) -> PartialOpticsViewInput:
    """Creates a view from the channel name

    Args:
        channel_name (str): The channel name

    Returns:
        View: The view
    """
    return PartialOpticsViewInput(
        objective=objective, camera=camera, instrument=instrument, **kwargs
    )


def timepoint_view_from_era(
    era: EraFragment,
    index_since_start: int = None,
    ms_since_start: Milliseconds = None,
    t: int = None,
) -> PartialTimepointViewInput:
    """Creates a view from the channel name

    Args:
        channel_name (str): The channel name

    Returns:
        View: The view
    """
    if t is not None:
        return PartialTimepointViewInput(
            era=era.id,
            index_since_start=index_since_start,
            ms_since_start=ms_since_start,
            tMax=t,
            tMin=t,
        )

    return PartialTimepointViewInput(
        era=era.id,
        index_since_start=index_since_start,
        ms_since_start=ms_since_start,
    )


def labelview_from_agents(
    fluorophore: Optional[FluorophoreFragment] = None,
    primary_antibody: AntibodyFragment = None,
    secondary_antibody: AntibodyFragment = None,
    c: int = None,
    **kwargs,
) -> PartialLabelViewInput:
    """Creates a view from the channel name

    Args:
        channel_name (str): The channel name

    Returns:
        View: The view
    """
    if c is not None:
        return PartialLabelViewInput(
            fluorophore=fluorophore.id,
            primary_antibody=primary_antibody.id,
            secondary_antibody=secondary_antibody.id,
            cMax=c,
            cMin=c,
            **kwargs,
        )

    return PartialLabelViewInput(
        fluorophore=fluorophore.id,
        primary_antibody=primary_antibody.id,
        secondary_antibody=secondary_antibody.id,
        **kwargs,
    )


def view_from_color(
    rgb_context: ID,
    c: int = None,
    r_scale: float = 1,
    g_scale: float = 1,
    b_scale: float = 1,
    **kwargs,
) -> PartialRGBViewInput:
    """Creates a view from the channel name

    Args:
        channel_name (str): The channel name

    Returns:
        View: The view
    """
    if c is not None:
        return PartialRGBViewInput(
            context=rgb_context,
            cMax=c,
            cMin=c,
            rScale=r_scale,
            gScale=g_scale,
            bScale=b_scale,
            **kwargs,
        )
